/*
 * LuxoCanvas.java
 *
 * Created on 11. Dezember 2007, 20:41
 */

/*
 The contents of this file are subject to the
 Mozilla Public License Version 1.1 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
 
 http://www.mozilla.org/MPL/
 
 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied.
 See the License for the specific language governing rights
 and limitations under the License.
 
 The Initial Developer of the Original Code is
 Georg-Johann Lay (GeorgJohann@web.de).
 
 Portions created by the Initial Developer are
 Copyright (C) 2007 Georg-Johann Lay. All Rights Reserved.
 
 Alternatively, the contents of this file may be used under the terms
 of either the GNU General Public License Version 3 or later (the "GPL"),
 or the GNU Lesser General Public License Version 3 or later (the "LGPL"),
 in which case the provisions of the GPL or the LGPL are applicable instead
 of those above.
 
 If you wish to allow use of your version of this file only
 under the terms of either the GPL or the LGPL, and not to allow others to
 use your version of this file under the terms of the MPL, indicate your
 decision by deleting the provisions above and replace them with the notice
 and other provisions required by the LGPL or the GPL. If you do not delete
 the provisions above, a recipient may use your version of this file under
 the terms of any one of the MPL, the GPL or the LGPL.
 */

package de.gjlay.mobile.luxo;

import javax.microedition.lcdui.Canvas;
import javax.microedition.lcdui.Display;
import javax.microedition.lcdui.Graphics;

/**
 * This is the working canvas that triggers the backlight (if platfrom allows that).
 * @author Georg-Johann Lay
 */
final class LuxoCanvas 
        extends Canvas 
        implements Runnable
{
    /** Index of red component */
    static final int RED   = 0;
    
    /** Index of green component */
    static final int GREEN = 1;
    
    /** Index of blue component */
    static final int BLUE  = 2;
    
    /** our midlet's dsplay */
    private Display display;

    /** full screen mode? */
    private boolean isFullScreen = true;
    
    /** trigger backlight (or be inactive)? */
    private boolean isOn = true;
    
    /** just after turning off or when pressed a kexy while 'off' */
    private boolean justTurnedOff = false;
    
    /** displayed color: reg, green, blue component */
    private int[] colorRGB = 
    {
        0xff, 0xff, 0xff
    };

    /** Cycle through colors? */
    private boolean rotate = false;
    
    /** Index used during color cycling */
    private int rotIndex = RED;
    
    /** displayed color during cycling: reg, green, blue component */
    private int[] rotRGB = 
    {
        0, 0, 0
    };
    
    /** help strings: each entry is a topic/explanation pair */
    static final String[][] HelpText = 
    {
        {"Luxo", "is just a Lamp that turns on the Backlight if it is allowed to do so."},
        {"0", "Toggle Backlight On/Off"},
        {"*", "Darker"},
        {"#", "Brighter"},
        {"1", "More Red"},
        {"2", "More Green"},
        {"3", "More Blue"},
        {"4", "Less Red"},
        {"5", "Less Green"},
        {"6", "Less Blue"},
        {"8", "Toggle full Screen On/Off"},
        {"9", "Toggle Color cycling On/Off"}
    };

    /**
     * Constructs Luxo and starts its thread
     * @param display the midlet's display
     */
    public LuxoCanvas (Display display)
    {
        this.display = display;
        setFullScreenMode (isFullScreen);
        
        (new Thread (this)).start();
    }
    
    /**
     * paint on the graphics object
     * @param g graphics to paint on
     */
    public void paint (Graphics g)
    {
        int color = GetColor (rotate ? rotRGB : colorRGB);
        
        g.setColor (color);
        g.fillRect (0, 0, getWidth(), getHeight());
        
        String text = isOn ? "Luxo" : "Luxo Off";
        
        if (!isFullScreen)
            setTitle (text);
        else
        {
            // if in fullscreen mode (i.e. no title displayed) 
            // then draw some text in the middle of the display
            int h = g.getFont().getHeight();
            int w = g.getFont().stringWidth (text);
            int px = (getWidth()-w)/2;
            int py = (getHeight()-h)/2;

            g.setColor (getContrastBlackOrWhite (color));
            g.drawString (text, px, py, Graphics.TOP|Graphics.LEFT);
        }
    }
    
    /**
     * Called when a key is help down. Most keys are forwarded to {@link #keyPressed(int)}.
     * @param keyCode code of key
     */
    protected void keyRepeated (int keyCode)
    {
        switch (keyCode)
        {
            case KEY_STAR:
            case KEY_POUND:
                
            case KEY_NUM1:
            case KEY_NUM2:
            case KEY_NUM3:
            case KEY_NUM4:
            case KEY_NUM5:
            case KEY_NUM6:
                keyPressed (keyCode);
                return;
        }
    }
    
    /**
     * Called when a key is pressed.
     * @param keyCode code of key
     */
    protected void keyPressed (int keyCode)
    {
        // toggle permanent backlight trigger on/off
        if (KEY_NUM0 == keyCode)
            isOn ^= true;
        
        // toggle fullscreen mode
        if (KEY_NUM8 == keyCode)
            setFullScreenMode (isFullScreen ^= true);
        
        justTurnedOff = !isOn;
        
        if (!isOn)
        {
            repaint();
            return;
        }
        
        // toggle rotate
        if (KEY_NUM9 == keyCode)
            if (rotate ^= true)
            {
                rotRGB[RED]   = colorRGB[RED];
                rotRGB[GREEN] = colorRGB[GREEN];
                rotRGB[BLUE]  = colorRGB[BLUE];
            }
                
        if (!rotate) switch (keyCode)
        {
            // darker
            case KEY_STAR:
                decrement (RED);
                decrement (GREEN);
                decrement (BLUE);
                break;
                
            //brighter
            case KEY_POUND:
                increment (RED);
                increment (GREEN);
                increment (BLUE);
                break;

            // red up/down    
            case KEY_NUM1:  increment (RED);      break;
            case KEY_NUM4:  decrement (RED);      break;
            
            // green up/down    
            case KEY_NUM2:  increment (GREEN);      break;
            case KEY_NUM5:  decrement (GREEN);      break;
            
            // blue up/down    
            case KEY_NUM3:  increment (BLUE);      break;
            case KEY_NUM6:  decrement (BLUE);      break;
        }
        
        repaint();
    }
    
    /**
     * triggererd when display size changed
     * @param width new width
     * @param height new height
     */
    protected void sizeChanged (int width, int height)
    {
        repaint();
    }
    
    /**
     * sample color
     * @return color
     * @param rgb reg, green, blue component
     */
    private static int GetColor (int[] rgb)
    {
        return GetColor (rgb[RED], rgb[GREEN], rgb[BLUE]);
    }
    
    /**
     * sample color
     * @param r red component
     * @param g green component
     * @param b blue component
     * @return color
     */
    private static int GetColor (int r, int g, int b)
    {
        return (r << 16) | (g << 8) | b;
    }
    
    /**
     * get contrasting color, one of black or white
     * @param color a color
     * @return black or white that contrasts good with <CODE>color</CODE>
     */
    private int getContrastBlackOrWhite (int color)
    {
        if (rotate)
            return color ^ 0xffffff;
        
        int b = color & 0xff;
        int g = (color >>  8) & 0xff;
        int r = (color >> 16) & 0xff;
        
        int bright = (55*g + 30*r + 15*b)/100 < 0x70 ? 0xff : 0;
        
        return GetColor (bright, bright, bright);
    }
    
    /**
     * increment component of colorRGB
     * 
     * @param index component to increment
     */
    private void increment (int index)
    {
        colorRGB[index] = Sat8 (colorRGB[index]+17);
    }
    
    /**
     * decrement component of colorRGB
     * 
     * @param index component to decrement
     */
    private void decrement (int index)
    {
        colorRGB[index] = Sat8 (colorRGB[index]-17);
    }
    
    /**
     * Set <code>rotRGB</code> to the next color, used when cycling through colors. 
     * Travels along the edges of the color cube that are not connected to
     * black or white, i.e. red-magenta-blue-cyan-green-yellow-red-...
     */
    private void nextColor ()
    {
        int v1 = rotRGB[rotIndex % 3];
        int v2 = Sat8 (v1 + (rotIndex % 2 == 0 ? 17 : -17));
        
        if (v1 == v2)
            // no change: goto next component
            rotIndex = (rotIndex+1) % (2*3);
        else
            // change: write back the component
            rotRGB[rotIndex % 3] = v2;
    }
    
    /**
     * unsigned 8-bit saturation
     * @param val value to be saturated
     * @return 8-bit unsigned saturated value
     */
    private static int Sat8 (int val)
    {
        return val < 0 ? 0 : val > 0xff ? 0xff : val;
    }
    
    /**
     * Luxo must implement a runnable because we have to trigger the backlight permanantly.
     * Display just offers to flash the backlight, there is no real way to turn it on or off.
     * This may work or may not work, depending on the platform and MIDlet implementation.
     * @see Display#flashBacklight(int)
     */
    public void run() 
    {
        while (true) 
        {
            try 
            {
                Thread.sleep (150); 
            } 
            catch (InterruptedException ex) 
            {
            }

            synchronized (this)
            {
                if (isOn)
                {
                    display.flashBacklight (200);
                    nextColor();
                }
                else
                    if (justTurnedOff)
                    {
                        justTurnedOff = false;
                        display.flashBacklight (10);
                    }
                repaint();
            }
        }
    }
}
